from django.db import models
from .crypto import Crypto
from .exceptions import EncryptedFieldException


class EncryptedMixin(models.Field):
    internal_type = "CharField"
    prepared_max_length = None

    def __init__(self, key=None, **kwargs):
        kwargs.setdefault('max_length', self.prepared_max_length)
        self.crypto = Crypto(key)
        super().__init__(**kwargs)

    def get_db_prep_value(self, value, connection, prepared=False):
        value = super().get_db_prep_value(value, connection, prepared)
        if value is not None:
            encrypted_text = self.crypto.encrypt(value)
            if self.max_length and len(encrypted_text) > self.max_length:
                raise EncryptedFieldException(
                    f"Field {self.name} max_length={self.max_length} encrypted_len={len(encrypted_text)}"
                )
            return encrypted_text
        return None

    def from_db_value(self, value, expression, connection, context):
        if value is not None:
            return self.to_python(self.crypto.decrypt(value))
        return None

    def get_internal_type(self):
        return self.internal_type


class EncryptedTextField(EncryptedMixin, models.TextField):
    internal_type = "TextField"


class EncryptedCharField(EncryptedMixin, models.CharField):
    prepared_max_length =255


class EncryptedEmailField(EncryptedMixin, models.EmailField):
    prepared_max_length = 254


class EncryptedIntegerField(EncryptedMixin, models.CharField):
    prepared_max_length = 64

    def to_python(self, value):
        if value is None:
            return value
        try:
            return int(value)
        except (TypeError, ValueError):
            raise exceptions.ValidationError(
                self.error_messages['invalid'],
                code='invalid',
                params={'value': value},
            )

    def check(self, **kwargs):
        return [
            *super(models.CharField, self).check(**kwargs),
        ]



from django.core.management.base import BaseCommand

from mirage.tools import Migrator


class Command(BaseCommand):
    help = "encrypt/decrypt data use Mirage ~"

    def add_arguments(self, parser):
        parser.add_argument("--app", type=str, required=True)
        parser.add_argument("--model", type=str, required=True)
        parser.add_argument("--field", type=str, required=True)
        parser.add_argument("--method", type=str, required=True)
        parser.add_argument("--offset", type=int, required=False)
        parser.add_argument("--total", type=int, required=False)
        parser.add_argument("--limit", type=int, required=False)
        parser.add_argument("--tofield", type=str, required=False)

    def handle(self, *args, **options):
        app = options['app']
        model = options['model']
        field = options['field']
        method = options['method']
        offset = options['offset'] or 0
        total = options['total']
        limit = options['limit'] or 1000
        tofield = options['tofield']

        if method == 'encrypt':
            Migrator(app, model, field).encrypt(offset=offset, total=total, limit=limit)
        elif method == 'decrypt':
            Migrator(app, model, field).decrypt(offset=offset, total=total, limit=limit)
        elif method == 'copy':
            assert tofield is not None
            Migrator(app, model, field, tofield=tofield).copy(offset=offset, total=total, limit=limit)
        elif method == 'decrypt_to':
            assert tofield is not None
            Migrator(app, model, field, tofield=tofield).decrypt_to(offset=offset, total=total, limit=limit)
        elif method == 'encrypt_to':
            assert tofield is not None
            Migrator(app, model, field, tofield=tofield).encrypt_to(offset=offset, total=total, limit=limit)
        else:
            return



import base64

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import padding
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes

from django.conf import settings
from django.utils.encoding import force_bytes, force_text


class Crypto:

    def __init__(self, key=None):
        if key is None:
            key = getattr(settings, "SECRET_KEY")
            assert len(key) >= 32, "settings.SECRET_KEY length must more than 32!"
            self.key = base64.urlsafe_b64encode(force_bytes(key))[:32]
        else:
            assert len(key) >= 32, "key length must more than 32!"
            self.key = key[:32]

    def encrypt(self, text):
        if text is None:
            return None
        try:
            self.try_decrypt(text)
            return text
        except Exception:
            return self.try_encrypt(text)

    def try_encrypt(self, text):
        encryptor = Cipher(algorithms.AES(self.key), modes.ECB(), default_backend()).encryptor()
        padder = padding.PKCS7(algorithms.AES(self.key).block_size).padder()
        padded_data = padder.update(force_bytes(text)) + padder.finalize()
        encrypted_text = encryptor.update(padded_data) + encryptor.finalize()
        return force_text(base64.urlsafe_b64encode(encrypted_text))

    def try_decrypt(self, encrypted_text):
        decryptor = Cipher(algorithms.AES(self.key), modes.ECB(), default_backend()).decryptor()
        padder = padding.PKCS7(algorithms.AES(self.key).block_size).unpadder()
        decrypted_text = decryptor.update(base64.urlsafe_b64decode(encrypted_text))
        unpadded_text = padder.update(decrypted_text) + padder.finalize()
        return force_text(unpadded_text)

    def decrypt(self, encrypted_text):
        if encrypted_text is None:
            return None
        try:
            return self.try_decrypt(encrypted_text)
        except Exception:
            return encrypted_text
